@@ -7,11 +7,11 @@
 #  the U.S. Government retains certain rights in this software.
 #  _________________________________________________________________________
 
-from copy import deepcopy
 import re
-from sys import exc_info
+from sys import exc_info, stdout
 from textwrap import wrap
 import logging
+import pickle
 
 import six
 from six.moves import xrange
@@ -64,6 +64,43 @@ def _strip_indentation(doc):
     return '\n'.join(lines)
 
 
+class _UnpickleableDomain(object):
+    def __init__(self, obj):
+        self._type = type(obj).__name__
+        self._name = obj.name(True)
+
+    def __call__(self, arg):
+        logging.error(
+"""%s %s was pickled with an unpicklable domain.
+    The domain was stripped and lost during the pickle process.  Setting
+    new values on the restored object cannot be mapped into the correct
+    domain.
+""" % ( self._type, self._name))
+        return arg
+
+def _picklable(field,obj):
+    try:
+        pickle.dumps(field)
+        return field
+    except:
+        # Contrary to the documentation, Python is not at all consistent
+        # with the exception that is raised when pickling an object
+        # fails:
+        #
+        #    Python 2.6 - 3.4:  pickle.PicklingError
+        #    Python 3.5 - 4.6:  AttributeError
+        #    Python 2.6 - 2.7 (cPickle):  TypeError
+        #
+        # What we are concerned about is masking things like recursion
+        # errors.  Unfortunately, Python is not quite consistent there,
+        # either: exceeding recursion depth raises a RuntimeError
+        # through 3.4, then switches to a RecursionError (a derivative
+        # of RuntimeError).
+        if isinstance(exc_info()[0], RuntimeError):
+            raise
+        return _UnpickleableDomain(obj)
+
+
 class ConfigBase(object):
     __slots__ = ('_parent', '_name', '_userSet', '_userAccessed', '_data',
                  '_default', '_domain', '_description', '_doc', '_visibility',
@@ -75,7 +112,7 @@ class ConfigBase(object):
     NoArgument = (None,)
 
     def __init__(self,
-                 default,
+                 default=None,
                  domain=None,
                  description=None,
                  doc=None,
@@ -93,8 +130,6 @@ class ConfigBase(object):
         self._visibility = visibility
         self._argparse = None
 
-        self.reset()
-
     def __getstate__(self):
         # Nominally, __getstate__() should return:
         #
@@ -116,6 +151,8 @@ class ConfigBase(object):
         else:
             state = super(ConfigBase, self).__getstate__()
         state.update((key, getattr(self, key)) for key in ConfigBase.__slots__)
+        state['_domain'] = _picklable(state['_domain'], self)
+        state['_parent'] = None
         return state
 
     def __setstate__(self, state):
@@ -126,10 +163,25 @@ class ConfigBase(object):
             object.__setattr__(self, key, val)
 
     def __call__(self, value=NoArgument):
-        ans = deepcopy(self)
-        ans.reset()
-        ans._parent = None
-        ans._name = None
+        ans = self.__class__()
+        ans._default     = self._default
+        ans._domain      = self._domain
+        ans._description = self._description
+        ans._doc         = self._doc
+        ans._visibility  = self._visibility
+        if self.__class__ is ConfigBlock:
+            ans._implicit_declaration = self._implicit_declaration
+            ans._implicit_domain = self._implicit_domain
+            for k in self._decl_order:
+                if k in self._declared:
+                    v = self._data[k]
+                    ans._data[k] = _tmp = v()
+                    ans._decl_order.append(k)
+                    ans._declared.add(k)
+                    _tmp._parent = ans
+                    _tmp._name = v._name
+        else:
+            ans.reset()
         if value is not ConfigBase.NoArgument:
             ans.set_value(value)
         return ans
@@ -297,13 +349,15 @@ group, subparser, or (subparser, group)."""
                         del parsed_args.__dict__[_dest]
         return parsed_args
 
-    def display(self, content_filter=None, indent_spacing=2):
+    def display(self, content_filter=None, indent_spacing=2, ostream=None):
         if content_filter not in ConfigBlock.content_filters:
             raise ValueError("unknown content filter '%s'; valid values are %s"
                              % (content_filter, ConfigBlock.content_filters))
 
         _blocks = []
-        os = six.StringIO()
+        if ostream is None:
+            ostream=stdout
+
         for level, value, obj in self._data_collector(0, ""):
             if content_filter == 'userdata' and not obj._userSet:
                 continue
@@ -312,9 +366,8 @@ group, subparser, or (subparser, group)."""
 
             for i, v in enumerate(_blocks):
                 if v is not None:
-                    os.write(v)
+                    ostream.write(v)
                     _blocks[i] = None
-        return os.getvalue()
 
     def generate_yaml_template(self, indent_spacing=2, width=78, visibility=0):
         minDocWidth = 20
@@ -425,15 +478,21 @@ group, subparser, or (subparser, group)."""
             elif item_start:
                 os.write(indent + item_start)
             _doc = obj._doc or obj._description or ""
-            if '\n ' in _doc:
-                doc_lines = (item_body % (_doc),)
-            else:
-                doc_lines = wrap(
-                    item_body % (_doc),
-                    width,
-                    initial_indent=indent + ' ' * indent_spacing,
-                    subsequent_indent=indent + ' ' * indent_spacing)
             if _doc:
+                _wrapLines = '\n ' not in _doc
+                if '%s' in item_body:
+                    _doc = item_body % (_doc,)
+                elif _doc:
+                    _doc = item_body
+                if _wrapLines:
+                    doc_lines = wrap(
+                        _doc,
+                        width,
+                        initial_indent=indent + ' ' * indent_spacing,
+                        subsequent_indent=indent + ' ' * indent_spacing)
+                else:
+                    doc_lines = (_doc,)
+                # Write things out
                 os.writelines('\n'.join(doc_lines))
                 if not doc_lines[-1].endswith("\n"):
                     os.write('\n')
@@ -467,6 +526,10 @@ group, subparser, or (subparser, group)."""
 
 class ConfigValue(ConfigBase):
 
+    def __init__(self, *args, **kwds):
+        ConfigBase.__init__(self, *args, **kwds)
+        self.reset()
+
     def value(self, accessValue=True):
         if accessValue:
             self._userAccessed = True
@@ -487,16 +550,46 @@ class ConfigValue(ConfigBase):
 
 class ConfigList(ConfigBase):
 
-    def __getitem__(self, key):
-        self._userAccessed = True
-        if isinstance(self._data[key], ConfigValue):
-            return self._data[key].value()
+    def __init__(self, *args, **kwds):
+        ConfigBase.__init__(self, *args, **kwds)
+        if self._domain is None:
+            self._domain = ConfigValue()
+        elif isinstance(self._domain, ConfigBase):
+            pass
         else:
-            return self._data[key]
+            self._domain = ConfigValue(None, domain=self._domain)
+        self.reset()
 
-    def get(self, key):
+
+    def __setstate__(self, state):
+        state = super(ConfigList, self).__setstate__(state)
+        for x in self._data:
+            x._parent = self
+
+    def __getitem__(self, key):
+        val = self._data[key]
         self._userAccessed = True
-        return self._data[key]
+        if isinstance(val, ConfigValue):
+            return val.value()
+        else:
+            return val
+
+    def get(self, key, default=ConfigBase.NoArgument):
+        # Note: get() is borrowed from ConfigBlock for cases where we
+        # want the raw stored object (and to aviod the implicit
+        # conversion of ConfigValue members to their stored data).
+        try:
+            val = self._data[key]
+            self._userAccessed = True
+            return val
+        except:
+            pass
+        if default is ConfigBase.NoArgument:
+            return None
+        if self._domain is not None:
+            return self._domain(default)
+        else:
+            return ConfigValue(default)
 
     def __setitem__(self, key, val):
         # Note: this will fail if the element doesn't exist in _data.
@@ -538,12 +631,14 @@ class ConfigList(ConfigBase):
 
     def reset(self):
         ConfigBase.reset(self)
-        # Because the base reset() calls set_value, which will recreate
-        # the list from scratch, I do not think that we need to
-        # explicitly call the reset() function on any newly-created
-        # entries:
-        #for val in self._data:
-        #    val.reset()
+        # Because the base reset() calls set_value, any deefault list
+        # entries will get their userSet flag set.  This is wrong, as
+        # reset() should conceptually reset teh object to it's default
+        # state (e.g., before the user ever had a chance to mess with
+        # things).  As the list could contain a ConfigBlock, this is a
+        # recursive operation to put the userSet values back.
+        for val in self.user_values():
+            val._userSet = False
 
     def append(self, value=ConfigBase.NoArgument):
         val = self._cast(value)
@@ -556,9 +651,9 @@ class ConfigList(ConfigBase):
         self._data[-1]._userSet = True
         self._userSet = True
 
-    #@deprecated
     def add(self, value=ConfigBase.NoArgument):
-        #logger.warning("ConfigList.add() has been deprecated.  Use append()")
+        logger.warning(
+            "DEPRECATED: ConfigList.add() has been deprecated.  Use append()")
         return self.append(value)
 
     def _data_collector(self, level, prefix, visibility=None, docMode=False):
@@ -615,13 +710,19 @@ class ConfigBlock(ConfigBase):
         self._data = {}
 
     def __getstate__(self):
-        ans = super(ConfigBlock, self).__getstate__()
-        for key in ConfigBlock.__slots__:
-            ans[key] = getattr(self, key)
-        return ans
+        state = super(ConfigBlock, self).__getstate__()
+        state.update((key, getattr(self, key)) for key in ConfigBlock.__slots__)
+        state['_implicit_domain'] = _picklable(state['_implicit_domain'], self)
+        return state
+
+    def __setstate__(self, state):
+        state = super(ConfigBlock, self).__setstate__(state)
+        for x in six.itervalues(self._data):
+            x._parent = self
 
     def __getitem__(self, key):
         self._userAccessed = True
+        key = str(key)
         if isinstance(self._data[key], ConfigValue):
             return self._data[key].value()
         else:
@@ -629,24 +730,36 @@ class ConfigBlock(ConfigBase):
 
     def get(self, key, default=ConfigBase.NoArgument):
         self._userAccessed = True
+        key = str(key)
         if key in self._data:
             return self._data[key]
         if default is ConfigBase.NoArgument:
-            raise KeyError("Key '%s' not found in ConfigBlock %s" %
-                           (key, self.name(True)))
-        return default
+            return None
+        if self._implicit_domain is not None:
+            return self._implicit_domain(default)
+        else:
+            return ConfigValue(default)
+
+    def setdefault(self, key, default=ConfigBase.NoArgument):
+        self._userAccessed = True
+        key = str(key)
+        if key in self._data:
+            return self._data[key]
+        if default is ConfigBase.NoArgument:
+            return self.add(key, None)
+        else:
+            return self.add(key, default)
 
     def __setitem__(self, key, val):
+        key = str(key)
         if key not in self._data:
-            if self._implicit_domain is None:
-                self.add(key, ConfigValue(val))
-            else:
-                self.add(key, self._implicit_domain(val))
+            self.add(key, val)
         else:
             self._data[key].set_value(val)
         #self._userAccessed = True
 
     def __contains__(self, key):
+        key = str(key)
         return key in self._data
 
     def __len__(self):
@@ -699,6 +812,7 @@ class ConfigBlock(ConfigBase):
         return list(self.iteritems())
 
     def _add(self, name, config):
+        name = str(name)
         if config._parent is not None:
             raise ValueError(
                 "config '%s' is already assigned to Config Block '%s'; "
@@ -728,7 +842,14 @@ class ConfigBlock(ConfigBase):
             raise ValueError("Key '%s' not defined in Config Block '%s'"
                              " and Block disallows implicit entries" %
                              (name, self.name(True)))
-        ans = self._add(name, config)
+
+        if self._implicit_domain is None:
+            if isinstance(config, ConfigBase):
+                ans = self._add(name, config)
+            else:
+                ans = self._add(name, ConfigValue(config))
+        else:
+            ans = self._add(name, self._implicit_domain(config))
         self._userSet = True
         return ans
 
@@ -746,8 +867,14 @@ class ConfigBlock(ConfigBase):
             raise ValueError("Expected dict value for %s.set_value, found %s" %
                              (self.name(True), type(value).__name__))
         _implicit = []
+        _decl_map = {}
         for key in value:
-            if key not in self._data:
+            if str(key) in self._data:
+                # str(key) may not be key... store the mapping so that
+                # when we later iterate over the _decl_order, we can map
+                # the local keys back to the incoming value keys.
+                _decl_map[str(key)] = key
+            else:
                 if self._implicit_declaration:
                     _implicit.append(key)
                 else:
@@ -765,16 +892,12 @@ class ConfigBlock(ConfigBase):
             # things are deterministic and in case a validation depends
             # on the order)
             for key in self._decl_order:
-                if key in value:
+                if key in _decl_map:
                     #print "Setting", key, " = ", value
-                    self._data[key].set_value(value[key])
+                    self._data[key].set_value(value[_decl_map[key]])
             # implicit data is declated at the end (in sorted order)
-            if self._implicit_domain is None:
-                for key in sorted(_implicit):
-                    self.add(key, ConfigValue(value[key]))
-            else:
-                for key in sorted(_implicit):
-                    self.add(key, self._implicit_domain(value[key]))
+            for key in sorted(_implicit):
+                self.add(key, value[key])
         except:
             self.reset()
             self.set_value(_old_data)
